#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#  Copyright (C) 2022 Robert Szabo.
#
#  This software can be used by anyone at no cost, however,
#  if you like using my software and can support - please
#  donate money to a children's hospital of your choice.
#  This program is free software: you can redistribute it
#  and/or modify it under the terms of the GNU General Public
#  License as published by the Free Software Foundation:
#  GNU GPLv3. You must include this entire text with your
#  distribution.
#  This program is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.
#  See the GNU General Public License for more details.
#

# Created by rszabo50 at 2022-01-28

PROGRAM_NAME = 'ucm'
PROGRAM_TITLE = 'Urwid rendered Connection Manager'
PROGRAM_VERSION = '0.0.14'
SCM_URL = 'https://github.com/rszabo50/ucm'

MAIN_PALETTE = [
    ('header', 'dark cyan', 'default', "default", "dark cyan", "default"),
    ("normal", "dark cyan", "default"),
    ("highlighted", "light green", "default"),
    ("reverse-hl", "default", "dark cyan"),
    ("filter", 'default,underline', 'default', 'default,underline', 'default,underline', '#88a'),
    ('black', 'white', 'black'),
    ('dark red', 'white', 'dark red'),
    ('dark green', 'white', 'dark green'),
    ('brown', 'white', 'brown'),
    ('dark blue', 'white', 'dark blue'),
    ('dark magenta', 'white', 'dark magenta'),
    ('dark cyan', 'white', 'dark cyan'),
    ('light gray', 'black', 'light gray'),
    ('dark gray', 'white', 'dark gray'),
    ('light red', 'white', 'light red'),
    ('light green', 'white', 'light green'),
    ('yellow', 'black', 'yellow'),
    ('light blue', 'white', 'light blue'),
    ('light magenta', 'white', 'light magenta'),
    ('light cyan', 'black', 'light cyan'),
    ('white', 'black', 'white'),
    ('border', 'dark cyan', 'default'),
    ('shadow', 'default', 'light gray'),
    ('selectable', 'default', 'dark cyan'),
    ('focus', 'default', 'dark cyan', 'bold'),
    ('focustext', 'light gray', 'dark blue'),
    ('button normal', 'light gray', 'dark blue', 'standout'),
    ('button select', 'white', 'dark green')
]

# vim: ts=4 sw=4 et



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#  Copyright (C) 2022 Robert Szabo.
#
#  This software can be used by anyone at no cost, however,
#  if you like using my software and can support - please
#  donate money to a children's hospital of your choice.
#  This program is free software: you can redistribute it
#  and/or modify it under the terms of the GNU General Public
#  License as published by the Free Software Foundation:
#  GNU GPLv3. You must include this entire text with your
#  distribution.
#  This program is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.
#  See the GNU General Public License for more details.
#

# Created by rszabo50 at 2022-02-01

import logging
import os
import socket
import subprocess
import traceback
import time
from shlex import split

from urwid import Text, raw_display, Pile, ListBox, SimpleListWalker
from ucm.Widgets import ListView, ListItem
from ucm.Registry import Registry
from ucm.UserConfig import UserConfig
from ucm.Dialogs import DialogDisplay
from ucm.constants import MAIN_PALETTE


def host_is_local(hostname, port=None):
    """returns True if the hostname points to the localhost, otherwise False."""
    if port is None:
        port = 22  # no port specified, lets just use the ssh port
    hostname = socket.getfqdn(hostname)
    if hostname in ("localhost", "0.0.0.0", "127.0.0.1", "docker-desktop"):
        return True
    localhost = socket.gethostname()
    localaddrs = socket.getaddrinfo(localhost, port)
    targetaddrs = socket.getaddrinfo(hostname, port)
    for (family, socktype, proto, canonname, sockaddr) in localaddrs:
        for (rfamily, rsocktype, rproto, rcanonname, rsockaddr) in targetaddrs:
            if rsockaddr[0] == sockaddr[0]:
                return True
    return False


def build_ssh_command(host: str):
    connections = list(filter(lambda k: k['name'] == host, UserConfig().ssh_connections))
    if len(connections) == 1:
        data = connections[0]
        user_at_host = data['address'] if 'user' not in data else f"{data['user']}@{data['address']}"
        ident = f"-i {data['identity_file']}" if 'identity_file' in data else ''
        port = f"-p {data['port']}" if 'port' in data else ''
        opts = f"{data['options']}" if 'options' in data else ''
        return f"ssh -t {ident} {port} {opts} {user_at_host}"


# noinspection PyStatementEffect
def swarm_connect(data: any, shell: str = 'bash'):
    command = f"docker exec -it  {data['container']} {shell}"
    if not host_is_local(data['host']):
        command = f"{build_ssh_command(data['host'])} {command}"

    if Registry().get('main_loop'):
        Registry().main_loop.screen.stop()
        print(chr(27) + "[2J")
        print(command)
        logging.info(command)
        rc = os.system(command)
        if rc == 32256:
            rc = os.system(command.replace(" bash", " sh"))
        if rc != 0:
            print(f"Return code: {rc}")
            time.sleep(5)
        Registry().main_loop.screen.start(alternate_buffer=True)
        Registry().main_loop.screen.clear()


# noinspection PyStatementEffect
def swarm_inspect(data: any):

    command = f"{UserConfig().docker} inspect {data['container']}"
    if not host_is_local(data['host']):
        command = f"{build_ssh_command(data['host'])} {command}"

    proc = subprocess.Popen(split(command), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return [Text(x.rstrip()) for x in proc.communicate()[0].splitlines()]


# noinspection PyStatementEffect
class SwarmListView(ListView):

    def __init__(self):
        super().__init__('Swarm', filter_fields=['stack', 'host', 'container'])

    def formatter(self, record: any):

        display_host = record['host'] if len(record['host']) <= 40 else f'...{record["host"][-37:]}'
        display_stack = record['stack'] if len(record['stack']) <= 20 else f'...{record["stack"][-17:]}'
        display_container = record['container'] if len(record['container']) <= 40 else f'{record["container"][0:37]}...'

        return f"{str(record['index']).ljust(3)}   {display_stack.ljust(20)}   {display_host.ljust(40)}   {display_container.ljust(40)}   {record['image']}"

    # noinspection PyMethodMayBeStatic
    def fetch_data(self):
        data = []
        # noinspection PyBroadException,PyPep8
        try:
            proc = subprocess.Popen(['bash', os.path.join(os.path.dirname(__file__), 'ucm-swarmlisting.sh')], stdout=subprocess.PIPE)
            while True:
                line = proc.stdout.readline()
                if not line:
                    break
                parts = line.decode('UTF-8').split()
                if 'CONTAINER' not in parts[0]:
                    data.append({'stack': parts[0].strip(), 'container': parts[1].strip(), 'host': parts[2].strip(), 'image': parts[3].strip()})
        except Exception as _e:
            logging.error(traceback.format_exc())

        return data

    def double_click_callback(self):
        logging.debug(f'{self.name}] {self.selected.item_data["container"]} double_click_callback')
        swarm_connect(self.selected.item_data)

    @staticmethod
    def popup_info_dialog(data):
        cols, rows = raw_display.Screen().get_cols_rows()
        d = DialogDisplay(f"Container Inpsection: {data['name']}", cols - 20, rows - 6,
                          body=Pile([
                              ListBox(
                                  SimpleListWalker(swarm_inspect(data))
                              )]),
                          loop=Registry().main_loop,
                          exit_cb=SwarmListView.close_cb,
                          palette=MAIN_PALETTE)
        d.add_buttons([("Cancel", 1)])
        d.show()

    @staticmethod
    def close_cb(button: any):
        pass

    def keypress_callback(self, size, key, list_item: ListItem = None):
        logging.debug(f'ListViewHandler[{self.name}] {size} {key} pressed')
        if key in ['c', 'b']:
            self.connect(list_item.item_data)
        elif key == 's':
            self.connect(list_item.item_data, shell='sh')

    def keypress_callback(self, size, key, data: any = None):
        logging.debug(f'ListViewHandler[{self.name}] {size} {key} pressed')
        if key == 'c':
            swarm_connect(data, shell='bash')
        if key == 'i':
            SwarmListView.popup_info_dialog(data)

    def get_header(self):
        return f"{'#'.ljust(3)}   {'Stack'.ljust(20)}   {'Host'.ljust(40)}   {'Container'.ljust(40)}   Image"

# vim: ts=4 sw=4 et



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#  Copyright (C) 2022 Robert Szabo.
#
#  This software can be used by anyone at no cost, however,
#  if you like using my software and can support - please
#  donate money to a children's hospital of your choice.
#  This program is free software: you can redistribute it
#  and/or modify it under the terms of the GNU General Public
#  License as published by the Free Software Foundation:
#  GNU GPLv3. You must include this entire text with your
#  distribution.
#  This program is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.
#  See the GNU General Public License for more details.
#

# Created by rszabo50 at 2022-01-28

from urwid import Frame, Button, Text, WidgetWrap, Divider, Pile, Padding, Filler, AttrWrap, LineBox, Columns, SolidFill, GridFlow, MainLoop, Overlay
from urwid import CENTER, MIDDLE
from ucm.Registry import Registry

import time


class DialogButton(Button):
    button_right = Text("")
    button_left = Text("")


# noinspection PyArgumentList
class DialogFrame(Frame):

    def keypress(self, size, key):
        if key == 'tab':
            if self.focus_part == 'body':
                self.set_focus('footer')
                return
            elif self.focus_part == 'footer':
                self.set_focus('body')
                return
        super().keypress(size, key)


class DialogDisplay(WidgetWrap):
    default_palette = [
        ('body', 'black', 'white'),
        ('border', 'black', 'white'),
        ('shadow', 'white', 'black'),
        ('selectable', 'black', 'dark cyan'),
        ('focus', 'black', 'dark cyan', 'bold'),
        ('focustext', 'light gray', 'dark blue'),
        ('button normal', 'light gray', 'dark blue', 'standout'),
        ('button select', 'white', 'dark green'),
    ]
    parent = None

    def __init__(self, text, width, height, body: any = None, loop: any = None, exit_cb: any = None,
                 palette: list = None):
        width = int(width) if width > 0 else ('relative', 80)
        height = int(height) if height > 0 else ('relative', 80)
        self.exit_cb = exit_cb
        self.buttons = []
        self.loop=loop
        self.palette = palette if palette is not None else self.default_palette

        if body is None:
            self.body = SolidFill(' ')
            fp = 'footer'
        else:
            self.body = body
            fp = 'body'

        self.frame = DialogFrame(self.body, focus_part=fp)
        if text is not None:
            self.frame.header = Pile([Text(text), Divider(u'\u2550')])
        w = self.frame

        # pad area around listbox
        w = Padding(w, ('fixed left', 2), ('fixed right', 2))
        w = Filler(w, ('fixed top', 1), ('fixed bottom', 1))
        w = AttrWrap(w, 'body')

        w = LineBox(w)

        # "shadow" effect
        w = Columns([w, ('fixed', 1, AttrWrap(
            Filler(Text(('border', ' ')), "top")
            , 'shadow'))])

        w = Frame(w, footer=AttrWrap(Text(('border', ' ')), 'shadow'))
        if self.loop is None:
            # this dialog is the main window
            # create outermost border area
            w = Padding(w, CENTER, width)
            w = Filler(w, MIDDLE, height)
            w = AttrWrap(w, 'border')
        else:
            # this dialog is a child window
            # overlay it over the parent window
            self.loop = loop
            self.parent = self.loop.widget
            w = Overlay(w, self.parent, CENTER, width + 2, MIDDLE, height + 2)
        self.view = w

        # Call WidgetWrap.__init__ to correctly initialize ourselves
        WidgetWrap.__init__(self, self.view)

    def add_buttons(self, buttons):
        button_list = []
        for name, exitcode in buttons:
            b = DialogButton(name, self.button_press)
            b.exitcode = exitcode
            b = AttrWrap(b, 'button normal', 'button select')
            button_list.append(b)
        self.buttons = GridFlow(button_list, 10, 3, 1, CENTER)
        self.frame.footer = Pile([Divider(u'\u2500'), self.buttons], focus_item=1)

    def button_press(self, button):
        if self.parent is not None:
            self.loop.widget = self.parent
        self.exit_cb(button)

    def show(self):
        if self.loop is None:
            self.loop = MainLoop(self.view, self.palette)
            exited = False
            while not exited:
                self.loop.run()
        else:
            self.loop.widget = self.view

# vim: ts=4 sw=4 et



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Copyright (C) 2022 Robert Szabo.
#
#  This software can be used by anyone at no cost, however,
#  if you like using my software and can support - please
#  donate money to a children's hospital of your choice.
#  This program is free software: you can redistribute it
#  and/or modify it under the terms of the GNU General Public
#  License as published by the Free Software Foundation:
#  GNU GPLv3. You must include this entire text with your
#  distribution.
#  This program is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.
#  See the GNU General Public License for more details.
#

# Created by rszabo50 at 2022-02-09

from pathlib import Path
import os
import traceback
import yaml
import getpass
import subprocess
from shutil import which


class UserConfig(dict):
    __instance = None

    def __new__(cls, *args):
        if cls.__instance is None:
            cls.__instance = dict.__new__(cls)
        return cls.__instance

    def set(self, k, v, overwrite=True):
        if overwrite or self.get(k) is None:
            setattr(self.__instance, k, v)

    def get(self, k):
        return getattr(self.__instance, k) if hasattr(self.__instance, k) else None

    def initialize(self):
        self.set('config_folder', f"{Path.home()}/.ucm", overwrite=False)
        self.set('ssh_config_file', f"{Path.home()}/.ucm/ssh_connections.yml", overwrite=False)
        setattr(self, 'docker', which('docker'))
        setattr(self, 'swarm_host', UserConfig.is_swarm_host())
        self.build_dot_ucm()
        self.load_ssh_config()

    def load_ssh_config(self):
        self.set('ssh_connections', UserConfig.load_yaml(self.get("ssh_config_file"), must_exist=True))

    @staticmethod
    def is_swarm_host():
        result = subprocess.check_output(['docker', 'info', '--format', '{{.Swarm.ControlAvailable}}'], stderr=subprocess.STDOUT)
        return result.splitlines()[0].decode("utf-8").strip().lower() == 'true'

    def build_dot_ucm(self):
        if not os.path.exists(self.get('config_folder')):
            os.makedirs(self.get('config_folder'))
        if not os.path.exists(self.get('ssh_config_file')):
            with open(self.get('ssh_config_file'), 'w') as f:
                f.write('\n'.join([
                    "# Each ssh connection is defined as:",
                    "# - name: hostname or identifier",
                    "#   address: ip or dns resolvable name",
                    "#   user: username to connect with",
                    "#   port: tcp port to use: default: 22",
                    "#   identity_file: <identity file to use if needed and its not your default key>",
                    "#   options: <any other ssh options, supported by the command line",
                    "#   category: <identifier ; makes filtering easier>",
                    "# Note: Because UCM will use the address variable to connect, common options like the identity file, port etc",
                    "#       can be done via your ~/.ssh/config e.g.",
                    "#         Host: 192.168.0.*",
                    "#             GSSAPIAuthentication no",
                    "#             StrictHostKeyChecking no",
                    "#             UserKnownHostsFile /dev/null",
                    "#             IdentityFile ~/.ssh/MyIdentifyFile.pem",
                    "",
                    " - name: localhost",
                    "   address: 127.0.0.1",
                    f"   user: {getpass.getuser()}",
                    "   category: sample",
                    ""
                ]))

    @staticmethod
    def load_yaml(filename: str, must_exist: bool = False):
        if os.path.exists(filename):
            with open(filename, "r") as stream:
                try:
                    return yaml.load(stream, Loader=yaml.SafeLoader)
                except yaml.YAMLError as _e:
                    traceback.print_exc()
                    raise RuntimeError(f"Unable to load {filename}.")
        else:
            if must_exist:
                raise RuntimeError(f"Unable to load {filename}. NOT FOUND")


UserConfig().initialize()

# vim: ts=4 sw=4 et
